using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;

using gudusoft.gsqlparser;
using gudusoft.gsqlparser.Units;
using System.Globalization;
using System.Threading;
using System.Reflection;

namespace Analyze_SP
{
    #region SqlFileList
    public class sqlFileList
    {

        public List<FileInfo> sqlfiles = new List<FileInfo>();

        public sqlFileList(string pattern)
        {
            if (pattern == null)
                return;
            string patternString = pattern.Replace('/', '\\');
            int index = patternString.LastIndexOf('\\') + 1;
            string parentPath = "";
            string filePattern = pattern;
            if (index > 0)
            {
                parentPath = patternString.Substring(0, index);
                filePattern = patternString.Substring(index, patternString.Length - parentPath.Length);
            }
            DirectoryInfo parent = new DirectoryInfo(parentPath);
            if (parent.Exists)
                traverseSqlFiles(parent, filePattern);
        }

        private void traverseSqlFiles(DirectoryInfo parent, string filePattern)
        {
            if (parent == null || !parent.Exists || filePattern == null) return;
            String[] splits = filePattern.Split('*');
            StringBuilder buffer = new StringBuilder();
            buffer.Append(@"^(?i)");
            for (int i = 0; i < splits.Length; i++)
            {
                buffer.Append(@splits[i]);
                if (i < splits.Length - 1)
                {
                    buffer.Append(".*");
                }
            }
            if (filePattern.EndsWith("*") && !buffer.ToString().EndsWith("*"))
                buffer.Append(".*");
            buffer.Append(@"$");
            string pattern = buffer.ToString();
            FileSystemInfo[] files = parent.GetFileSystemInfos();
            for (int i = 0; i < files.Length; i++)
            {
                if (files[i] is FileInfo && Regex.IsMatch(files[i].Name, pattern))
                {
                    sqlfiles.Add((FileInfo)files[i]);
                }
            }
        }
    }
    #endregion

    #region spInfo
    public class spInfo
    {
        public string file;
        public string db;
        public List<procedureInfo> procedures = new List<procedureInfo>();
        public List<builtInFunctionInfo> functions = new List<builtInFunctionInfo>();
    }

    #endregion

    #region processInfo

    public enum objectType { SP, Table };
    public enum usageType { Exec, Read, Insert, Update, Create, Delete, Drop, Set, IfElse };

    public class procedureInfo
    {
        public string name;
        public bool hasTryCatch;
        public objectType objectType = objectType.Table;
        public List<operateInfo> operates = new List<operateInfo>();
    }
    #endregion

    #region operateInfo
    public class operateInfo
    {
        public string objectUsed;
        public objectType objectType;
        public usageType usageType;
        public List<string> columns = new List<string>();
    }
    #endregion


    #region tableInfo
    class tableInfo
    {
        public TSourceToken schema, table;

        public String AsText
        {
            get { return (schema == null ? "" : schema.AsText.Trim()) + (table == null ? "" : table.AsText.Trim()); }
        }
    }

    #endregion

    #region columnInfo
    class columnInfo
    {
        public tableInfo table;
        public TSourceToken column;
        public String AsText
        {
            get { return column == null ? "" : column.AsText.Trim(); }
        }
    }
    #endregion

    #region builtInFunctionInfo
    public class builtInFunctionInfo
    {
        public String function;
        public long lineNo, columnNo;
        public usageType stmtType;
    }
    #endregion

    public class analyze_sp
    {
        static void Main(string[] args)
        {
            Thread.CurrentThread.CurrentCulture = new CultureInfo("en-US", false);
            if (args.Length == 0)
            {
                Console.WriteLine("Usage: {0} scriptfile [/o <output file path>] [/d <csv delimiter character>]", Environment.GetCommandLineArgs()[0]);
                Console.WriteLine("/o: Option, write the output stream to the specified file.");
                Console.WriteLine("/d: Option, set the csv delimiter character. The default delimiter character is '|'");
                Console.WriteLine("/a: Option, check all items.");
                Console.WriteLine("/r: Option, check the database object relations in the store procedure.");
                Console.WriteLine("/f: Option, check the built-in functions in the store procedure.");
                Console.WriteLine("/t: Option, check if the store procedure contains try catch clause.");
                //Console.Read();
                return;
            }


            Array array = Array.CreateInstance(typeof(string), args.Length);

            for (int i = 0; i < array.Length; i++)
            {
                array.SetValue(args[i], i);
            }

            string outputFile = null;

            int index = Array.IndexOf(array, "/o");

            if (index != -1 && args.Length > index + 1)
            {
                outputFile = args[index + 1];
            }

            string delimiter = "|";

            index = Array.IndexOf(array, "/d");

            if (index != -1 && args.Length > index + 1)
            {
                delimiter = args[index + 1];
            }

            bool hasTryCatchClause = false;
            bool checkBuiltInFunction = false;
            bool checkDBObjectRelations = false;

            if (Array.IndexOf(array, "/a") != -1)
            {
                hasTryCatchClause = true;
                checkBuiltInFunction = true;
                checkDBObjectRelations = true;
            }
            else
            {
                if (Array.IndexOf(array, "/f") != -1)
                {
                    checkBuiltInFunction = true;
                }
                if (Array.IndexOf(array, "/t") != -1)
                {
                    hasTryCatchClause = true;
                }
                if (Array.IndexOf(array, "/r") != -1)
                {
                    checkDBObjectRelations = true;
                }
            }

            if (checkBuiltInFunction == false && hasTryCatchClause == false)
            {
                checkDBObjectRelations = true;
            }

            analyze_sp impact = new analyze_sp(args[0], delimiter);
            impact.setCheckBuiltInFunction(checkBuiltInFunction);
            impact.setCheckTryCatchClause(hasTryCatchClause);
            impact.setCheckDBObjectRelation(checkDBObjectRelations);
            impact.analyzeSQL();

            StreamWriter writer = null;
            if (outputFile != null)
            {
                try
                {
                    writer = new StreamWriter(outputFile);
                    Console.SetOut(writer);
                }
                finally { }
            }

            if (impact.checkDBObjectRelation)
            {
                Console.WriteLine("DB of Anayzed Object"
                        + delimiter
                        + "Name of Analyzed Object"
                        + delimiter
                        + "Object Type"
                        + delimiter
                        + "Object Used"
                        + delimiter
                        + "Object Type"
                        + delimiter
                        + "Usage Type"
                        + delimiter
                        + "Columns");
                Console.WriteLine(impact.DBObjectRelationsAnalysisResult);
            }
            if (impact.checkBuiltInFunction)
            {
                Console.WriteLine("File Name"
                        + delimiter
                        + "Built-in Function"
                        + delimiter
                        + "Line Number"
                        + delimiter
                        + "Column Number"
                        + delimiter
                        + "Usage Type"
                        + delimiter);
                Console.WriteLine(impact.BuiltInFunctionAnalysisResult);
            }
            if (impact.hasTryCatchClause)
            {
                Console.WriteLine("File Name"
                        + delimiter
                        + "DB of Anayzed Object"
                        + delimiter
                        + "Procedure"
                        + delimiter
                        + "With Try Catch");
                Console.WriteLine(impact.TryCatchClauseAnalysisResult);
            }

            if (outputFile == null)
            {
                Console.Read();
            }
            else
            {
                if (writer != null)
                {
                    writer.Close();
                }
            }
        }



        StringBuilder buffer = new StringBuilder();
        Hashtable spInfoMap = new Hashtable();
        List<FileInfo> files = new List<FileInfo>();
        private StringBuilder relationBuffer = new StringBuilder();
        private StringBuilder functionBuffer = new StringBuilder();
        private StringBuilder tryCatchBuffer = new StringBuilder();
        private bool checkBuiltInFunction, hasTryCatchClause,
                checkDBObjectRelation;
        private String delimiter;

        public void setCheckBuiltInFunction(bool checkBuiltInFunction)
        {
            this.checkBuiltInFunction = checkBuiltInFunction;
        }

        public void setCheckDBObjectRelation(bool checkObjectRelation)
        {
            this.checkDBObjectRelation = checkObjectRelation;
        }

        public void setCheckTryCatchClause(bool hasTryCatchClause)
        {
            this.hasTryCatchClause = hasTryCatchClause;
        }

        public String DBObjectRelationsAnalysisResult
        {
            get { return this.getDBObjectRelationsAnalysisResult(); }
        }

        public String BuiltInFunctionAnalysisResult
        {
            get { return this.getBuiltInFunctionAnalysisResult(); }
        }

        public String TryCatchClauseAnalysisResult
        {
            get { return this.getTryCatchClauseAnalysisResult(); }
        }

        public analyze_sp(string sqlFile, string delimiter)
        {

            sqlFileList sqlFiles = new sqlFileList(sqlFile);
            this.delimiter = delimiter;
            if (sqlFiles.sqlfiles.Count > 0)
            {
                for (int i = 0; i < sqlFiles.sqlfiles.Count; i++)
                {
                    files.Add(sqlFiles.sqlfiles[i]);
                    spInfo sp = new spInfo();
                    sp.file = sqlFiles.sqlfiles[i].Name;
                    spInfoMap.Add(sqlFiles.sqlfiles[i].Name, sp);
                }
            }

            if (sqlFiles.sqlfiles.Count > 0)
            {
                foreach (FileInfo file in files)
                {
                    spInfo spInfo = (spInfo)spInfoMap[file.Name];
                    foreach (procedureInfo procedure in spInfo.procedures)
                    {
                        foreach (operateInfo info in procedure.operates)
                        {
                            StringBuilder builder = new StringBuilder();
                            for (int i = 0; i < info.columns.Count; i++)
                            {
                                builder.Append(info.columns[i]);
                                if (i < info.columns.Count - 1)
                                {
                                    builder.Append(",");
                                }
                            }
                            buffer.Append(spInfo.db).Append(delimiter)
                                   .Append(procedure.name).Append(delimiter)
                                   .Append(procedure.objectType).Append(delimiter)
                                   .Append(info.objectUsed).Append(delimiter)
                                   .Append(info.objectType).Append(delimiter)
                                   .Append(info.usageType).Append(delimiter)
                                   .Append(builder).AppendLine();
                        }

                    }
                }
            }
        }

        private String getDBObjectRelationsAnalysisResult()
        {
            if (relationBuffer.Length == 0 && files != null)
            {
                foreach (FileInfo file in files)
                {
                    spInfo spInfo = (spInfo)spInfoMap[file.Name];
                    foreach (procedureInfo procedure in spInfo.procedures)
                    {
                        foreach (operateInfo info in procedure.operates)
                        {
                            StringBuilder builder = new StringBuilder();
                            for (int i = 0; i < info.columns.Count; i++)
                            {
                                builder.Append(info.columns[i]);
                                if (i < info.columns.Count - 1)
                                {
                                    builder.Append(",");
                                }
                            }
                            relationBuffer.Append(spInfo.db)
                                    .Append(delimiter)
                                    .Append(procedure.name)
                                    .Append(delimiter)
                                    .Append(procedure.objectType)
                                    .Append(delimiter)
                                    .Append(info.objectUsed)
                                    .Append(delimiter)
                                    .Append(info.objectType)
                                    .Append(delimiter)
                                    .Append(info.usageType)
                                    .Append(delimiter)
                                    .Append(builder)
                                    .Append("\r\n");
                        }

                    }
                }
            }
            return relationBuffer.ToString();
        }

        private String getTryCatchClauseAnalysisResult()
        {
            if (tryCatchBuffer.Length == 0 && files != null)
            {
                foreach (FileInfo file in files)
                {
                    spInfo spInfo = (spInfo)spInfoMap[file.Name];
                    foreach (procedureInfo procedure in spInfo.procedures)
                    {
                        tryCatchBuffer.Append(file.Name)
                                .Append(delimiter)
                                .Append(spInfo.db)
                                .Append(delimiter)
                                .Append(procedure.name)
                                .Append(delimiter)
                                .Append(procedure.hasTryCatch ? "Yes" : "No")
                                .Append("\r\n");
                    }
                }
            }
            return tryCatchBuffer.ToString();
        }

        private String getBuiltInFunctionAnalysisResult()
        {
            if (functionBuffer.Length == 0 && files != null)
            {
                foreach (FileInfo file in files)
                {
                    spInfo spInfo = (spInfo)spInfoMap[file.Name];
                    foreach (builtInFunctionInfo function in spInfo.functions)
                    {
                        functionBuffer.Append(file.Name)
                                .Append(delimiter)
                                .Append(function.function)
                                .Append(delimiter)
                                .Append(function.lineNo)
                                .Append(delimiter)
                                .Append(function.columnNo)
                                .Append(delimiter)
                                .Append(function.stmtType)
                                .Append("\r\n");
                    }
                }
            }
            return functionBuffer.ToString();
        }

        public void analyzeSQL()
        {
            foreach (FileInfo file in files)
            {
                TGSqlParser sqlparser = new TGSqlParser(TDbVendor.DbVMssql);
                sqlparser.Sqlfilename = file.FullName;
                int ret = sqlparser.Parse();
                if (ret != 0)
                {
                    Console.WriteLine("Parse file " + sqlparser.Sqlfilename + " failed.");
                    Console.WriteLine(sqlparser.ErrorMessages);
                    continue;
                }
                spInfo spInfo = (spInfo)spInfoMap[file.Name];
                analyzeSQL(spInfo, sqlparser);
            }
        }

        protected void analyzeSQL(spInfo spInfo, TGSqlParser sqlparser)
        {
            procedureInfo procedureInfo = null;
            for (int i = 0; i < sqlparser.SqlStatements.Count(); i++)
            {
                TCustomSqlStatement sql = sqlparser.SqlStatements[i];
                if (sql is TMssqlUse)
                {
                    spInfo.db = ((TMssqlUse)sql).dbname.AsText;
                    procedureInfo = new procedureInfo();
                    spInfo.procedures.Add(procedureInfo);
                }
                else if (sql is TMssqlCreateProcedure)
                {
                    procedureInfo.name = ((TMssqlCreateProcedure)sql).procedurename.AsText;
                    procedureInfo.objectType = objectType.SP;
                    if (checkDBObjectRelation)
                    {
                        analyzeProcedure(procedureInfo, (TMssqlCreateProcedure)sql);
                    }
                    if (hasTryCatchClause)
                    {
                        checkTryCatchClause(procedureInfo,
                                sql.SourceTokenList);
                    }
                }
                else if (procedureInfo != null)
                {
                    analyzeSqlStatement(procedureInfo, sql);
                }
                if (checkBuiltInFunction)
                {
                    checkFunction(spInfo, sql.SourceTokenList);
                }
            }
        }

        void analyzeProcedure(procedureInfo procedureInfo, TMssqlCreateProcedure procedure)
        {
            for (int i = 0; i < procedure.SqlStatements.Count(); i++)
            {
                TCustomSqlStatement stmt = procedure.SqlStatements[i];
                analyzeSqlStatement(procedureInfo, stmt);
            }
        }

        private void analyzeSqlStatement(procedureInfo procedureInfo, TCustomSqlStatement stmt)
        {
            if (stmt is TMssqlBlock)
            {
                TMssqlBlock block = (TMssqlBlock)stmt;
                if (block.SqlStatements != null)
                {
                    for (int i = 0; i < block.SqlStatements.Count(); i++)
                    {
                        analyzeSqlStatement(procedureInfo, block.SqlStatements[i]);
                    }
                }
            }
            else if (stmt is TMssqlIfElse)
            {
                TMssqlIfElse ifElse = (TMssqlIfElse)stmt;
                if (ifElse.ifstmt != null)
                {
                    analyzeSqlStatement(procedureInfo, ifElse.ifstmt._sqlstmt);
                }
                if (ifElse.condition != null)
                {

                }
                if (ifElse.elsestmt != null)
                {
                    analyzeSqlStatement(procedureInfo, ifElse.elsestmt._sqlstmt);
                }
            }
            else if (stmt is TMssqlDeclare)
            {
                TMssqlDeclare declareStmt = (TMssqlDeclare)stmt;
                if (declareStmt.selectstmt != null && declareStmt.selectstmt.AsText.Trim().Length > 0)
                {
                    analyzeSqlStatement(procedureInfo, declareStmt.selectstmt);
                }
            }
            else if (stmt is TMssqlExecute && ((TMssqlExecute)stmt).ModuleName != null)
            {
                TMssqlExecute executeStmt = (TMssqlExecute)stmt;
                operateInfo operateInfo = new operateInfo();
                operateInfo.objectType = objectType.SP;
                operateInfo.objectUsed = executeStmt.ModuleName.AsText.Trim();
                operateInfo.usageType = usageType.Exec;
                procedureInfo.operates.Add(operateInfo);
            }
            else if (stmt is TCreateTableSqlStatement)
            {
                TCreateTableSqlStatement createStmt = (TCreateTableSqlStatement)stmt;
                TLzFieldList columns = createStmt.Table.Fields;
                operateInfo operateInfo = new operateInfo();
                operateInfo.objectType = objectType.Table;
                operateInfo.objectUsed = createStmt.Table.AsText.Trim();
                operateInfo.usageType = usageType.Create;
                foreach (TLzField column in columns)
                {
                    operateInfo.columns.Add(column.Name);
                }
                procedureInfo.operates.Add(operateInfo);

            }
            else if (stmt is TInsertSqlStatement)
            {
                TInsertSqlStatement insertStmt = (TInsertSqlStatement)stmt;
                TLzFieldList columns = insertStmt.Fields;
                operateInfo operateInfo = new operateInfo();
                operateInfo.objectType = objectType.Table;
                operateInfo.objectUsed = insertStmt.Table.AsText.Trim();
                operateInfo.usageType = usageType.Insert;
                foreach (TLzField column in columns)
                {
                    operateInfo.columns.Add(column.Name);
                }
                procedureInfo.operates.Add(operateInfo);

                if (insertStmt.ExecStmt != null)
                {
                    analyzeSqlStatement(procedureInfo, insertStmt.ExecStmt);
                }
            }
            else if (stmt is TUpdateSqlStatement)
            {
                TUpdateSqlStatement updateStmt = (TUpdateSqlStatement)stmt;
                TLzFieldList columns = updateStmt.Fields;
                operateInfo operateInfo = new operateInfo();
                operateInfo.objectType = objectType.Table;
                operateInfo.objectUsed = updateStmt.Table.AsText.Trim();
                operateInfo.usageType = usageType.Update;
                foreach (TLzField column in columns)
                {
                    operateInfo.columns.Add(column.FieldExpr.lexpr.AsText.Trim());
                }
                procedureInfo.operates.Add(operateInfo);
            }
            else if (stmt is TDeleteSqlStatement)
            {
                TDeleteSqlStatement deleteStmt = (TDeleteSqlStatement)stmt;
                operateInfo operateInfo = new operateInfo();
                operateInfo.objectType = objectType.Table;
                operateInfo.objectUsed = deleteStmt.TableText;
                operateInfo.usageType = usageType.Delete;
                procedureInfo.operates.Add(operateInfo);
            }
            else if (stmt is TMssqlDropTable)
            {
                TMssqlDropTable dropStmt = (TMssqlDropTable)stmt;
                operateInfo operateInfo = new operateInfo();
                operateInfo.objectType = objectType.Table;
                operateInfo.objectUsed = dropStmt.ndTable.AsText.Trim();
                operateInfo.usageType = usageType.Drop;
                procedureInfo.operates.Add(operateInfo);
            }
            else if (stmt is TSelectSqlStatement)
            {
                TSelectSqlStatement selectStmt = (TSelectSqlStatement)stmt;
                List<columnInfo> columnInfos = new List<columnInfo>();
                List<tableInfo> tableInfos = new List<tableInfo>();
                tableTokensInStmt(columnInfos, tableInfos, selectStmt);
                for (int j = 0; j < selectStmt.ChildNodes.Count(); j++)
                {
                    if (selectStmt.ChildNodes[j] is TCustomSqlStatement)
                    {
                        tablesInStmt(columnInfos, tableInfos, selectStmt.ChildNodes[j] as TCustomSqlStatement);
                    }
                }
                Hashtable columnMap = new Hashtable();
                for (int i = 0; i < columnInfos.Count; i++)
                {
                    columnInfo column = columnInfos[i];
                    tableInfo table = column.table;
                    if (columnMap.ContainsKey(table))
                    {
                        List<columnInfo> columns = (List<columnInfo>)columnMap[table];
                        bool flag = false;
                        foreach (columnInfo temp in columns)
                        {
                            if (temp.AsText.Equals(column.AsText, StringComparison.OrdinalIgnoreCase))
                            {
                                flag = true;
                                break;
                            }
                        }
                        if (!flag)
                        {
                            columns.Add(column);
                        }
                    }
                    else
                    {
                        List<columnInfo> columns = new List<columnInfo>();
                        columnMap.Add(table, columns);
                        columns.Add(column);
                    }
                }
                for (int i = 0; i < tableInfos.Count; i++)
                {
                    operateInfo operateInfo = new operateInfo();
                    operateInfo.objectType = objectType.Table;
                    operateInfo.objectUsed = tableInfos[i].AsText;
                    if (tableInfos[i].table.Location == TLzOwnerLocation.elintoclause)
                        operateInfo.usageType = usageType.Insert;
                    else
                        operateInfo.usageType = usageType.Read;
                    if (columnMap.ContainsKey(tableInfos[i]))
                    {
                        foreach (columnInfo column in (List<columnInfo>)columnMap[tableInfos[i]])
                        {
                            operateInfo.columns.Add(column.AsText);
                            operateInfo.objectUsed = column.table.AsText;
                        }
                    }
                    procedureInfo.operates.Add(operateInfo);
                }
            }
        }

        protected void checkFunction(spInfo spInfo, TSourceTokenList tokenList)
        {
            for (int i = 0; i < tokenList.Count(); i++)
            {
                TSourceToken token = tokenList[i];
                if (token.DBObjType == TDBObjType.ttObjFunction)
                {
                    Object stmt = null;
                    bool isFunc = false;
                    if (token.ParentAttr != null)
                    {
                        FieldInfo field1 = token.ParentAttr.GetType().GetField("DoParseSender", BindingFlags.NonPublic | BindingFlags.Instance);
                        FieldInfo field2 = token.ParentAttr.GetType().GetField("doParseSql", BindingFlags.NonPublic | BindingFlags.Instance);
                        if (field1 != null && field2 != null)
                        {
                            Object sender = field1.GetValue(token.ParentAttr);
                            Object sql = field2.GetValue(token.ParentAttr);
                            if (sender is TLz_FuncCall && sql != null)
                            {
                                isFunc = true;
                                stmt = sql;
                            }
                        }

                    }

                    if (!isFunc)
                        continue;

                    builtInFunctionInfo function = new builtInFunctionInfo();
                    function.function = token.AsText;
                    function.lineNo = token.YPosition;
                    function.columnNo = token.XPosition;

                    if (stmt is TInsertSqlStatement)
                    {
                        function.stmtType = usageType.Insert;
                    }
                    else if (stmt is TSelectSqlStatement)
                    {
                        function.stmtType = usageType.Read;
                    }
                    else if (stmt is TUpdateSqlStatement)
                    {
                        function.stmtType = usageType.Update;
                    }
                    else if (stmt is TDeleteSqlStatement)
                    {
                        function.stmtType = usageType.Delete;
                    }
                    else if (stmt is TMssqlDropTable)
                    {
                        function.stmtType = usageType.Drop;
                    }
                    else if (stmt is TDropTableSqlStatement)
                    {
                        function.stmtType = usageType.Drop;
                    }
                    else if (stmt is TMssqlExecute)
                    {
                        function.stmtType = usageType.Exec;
                    }
                    else if (stmt is TMssqlSet)
                    {
                        function.stmtType = usageType.Set;
                    }
                    else if (stmt is TMssqlCreateProcedure)
                    {
                        function.stmtType = usageType.Create;
                    }
                    else if (stmt is TMssqlIfElse)
                    {
                        function.stmtType = usageType.IfElse;
                    }
                    else
                    {
                        //function.stmtType = usageType.Create;
                    }
                    spInfo.functions.Add(function);
                }
            }
        }

        protected void checkTryCatchClause(procedureInfo procedureInfo,
                TSourceTokenList tokenList)
        {
            for (int i = 0; i < tokenList.Count(); i++)
            {
                TSourceToken token = tokenList[i];
                if (token.AsText.Trim().Equals("try", StringComparison.OrdinalIgnoreCase) &&
                    token.WordType == TWordType.wtReservedWord)
                {
                    procedureInfo.hasTryCatch = true;
                }
            }
        }

        void tableTokensInStmt(List<columnInfo> columnInfos, List<tableInfo> tableInfos, TCustomSqlStatement stmt)
        {
            TSourceTokenList tokenList = stmt.TableTokens;
            TSourceToken st = null;
            for (int i = 0; i < tokenList.Count(); i++)
            {
                tableInfo tableInfo = new tableInfo();
                st = tokenList[i];
                tableInfo.table = st;
                if (st.ParentToken != null)
                {
                    tableInfo.schema = st.ParentToken;
                }
                tableInfos.Add(tableInfo);
                columnsInTableToken(columnInfos, stmt, tableInfo);
            }
        }

        void columnsInTableToken(List<columnInfo> columnInfos, TCustomSqlStatement stmt, tableInfo tableInfo)
        {
            TSourceToken st = tableInfo.table;
            if (st.RelatedToken != null)
            {
                // declared table alias token
                TSourceToken rt = st.RelatedToken;
                TSourceToken rrt = null;
                for (int i = 0; i < rt.RelatedTokens.Count(); i++)
                {
                    rrt = rt.RelatedTokens[i];
                    if (rrt.ChildToken != null)
                    {
                        columnInfo columnInfo = new columnInfo();
                        columnInfo.table = tableInfo;
                        columnInfo.column = rrt.ChildToken;
                        columnInfos.Add(columnInfo);
                    }
                }
            }

            TSourceToken rtt = null;
            for (int i = 0; i < st.RelatedTokens.Count(); i++)
            {
                // reference table token
                rtt = st.RelatedTokens[i];
                if (rtt.DBObjType == TDBObjType.ttObjField)
                {
                    // get all field tokens link with table token (those token not linked by syntax like tablename.fieldname)
                    // but like this : select f from t

                    columnInfo columnInfo = new columnInfo();
                    columnInfo.table = tableInfo;
                    columnInfo.column = rtt;
                    columnInfos.Add(columnInfo);
                }
                if (rtt.ChildToken != null)
                {
                    columnInfo columnInfo = new columnInfo();
                    columnInfo.table = tableInfo;
                    columnInfo.column = rtt.ChildToken;
                    columnInfos.Add(columnInfo);
                }
            }
        }


        void tablesInStmt(List<columnInfo> columnInfos, List<tableInfo> tableInfos, TCustomSqlStatement stmt)
        {

            tableTokensInStmt(columnInfos, tableInfos, stmt);

            for (int j = 0; j < stmt.ChildNodes.Count(); j++)
            {
                if (stmt.ChildNodes[j] is TCustomSqlStatement)
                {
                    tablesInStmt(columnInfos, tableInfos, stmt.ChildNodes[j] as TCustomSqlStatement);
                }
            }

        }

    }
}
